/* execvp.c */

/* This and the other exec*.c files in this directory require 
   the target to provide the _execve syscall.  */

#include <_ansi.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#if defined (_WIN32)
#define PATH_DELIM ';'
#else
#define PATH_DELIM ':'
#endif

/*
 * Copy string, until c or <nul> is encountered.
 * NUL-terminate the destination string (s1).
 */

static char *
_DEFUN (strccpy, (s1, s2, c),
	char *s1 _AND
	char *s2 _AND
	char c)
{
  char *dest = s1;

  while (*s2 && *s2 != c)
    *s1++ = *s2++;
  *s1 = 0;

  return dest;
}

int
_DEFUN (execvp, (file, argv),
	_CONST char *file _AND
	char * _CONST argv[])
{
  char *path = getenv ("PATH");
  char buf[MAXNAMLEN];

  if (file[0] == '/')		/* pathname -- easy out */
    return execv (file, argv);

#ifdef _WIN32			/* Ditto for DOS style paths.  */
  if (isalpha (file[0]) && file[1] == ':')
    return execv (file, argv);
#endif

  buf[0] = 0;

  if (!path)
    return execv (file, argv);

  while (*path)
    {
      strccpy (buf, path, PATH_DELIM);
      strcat (buf, "/");
      strcat (buf, file);
      if (execv (buf, argv) == -1 && errno != ENOENT)
	return -1;
      while (*path && *path != PATH_DELIM)
	path++;
      if (*path == PATH_DELIM)
	path++;			/* skip over delim */
    }

  return -1;
}

